;; Programa: Package para entorno interface funcional de
;;           comunicaciones con agentes de juegos,
;;           utilizado para el juego PlanTron.
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    12/04/2006
;; Comentarios:
;;           Este package es el encargado de comunicar a los agentes
;;           jugadores, visualizadores y/o logs con el servidor
;;           o controlador de partidas PT-BASE y/o con otros
;;           agentes.
;;
;;           Esta implementación utiliza para dicha comunicacioón
;;           llamadas a funciones, esto es, emite información
;;           llamando a funciones exportadas en el servidor del
;;           juego (PT-BASE) y en los agentes, y recibe información
;;           mediante llamadas a sus funciones que exporta.
;;           El hacerlo  de esta forma limita a que el servidor
;;           del juego, este package y todos los agentes estén
;;           ejecutandose en la mísma máquina y en la misma
;;           instaciación del LISP.
;;
;;           Esta interface de comunicación aisla también al
;;           servidor del juego de los agentes, de forma que
;;           el servidor no tenga por qué saber que agentes
;;           están jugando (nombres, características, funciones,
;;           descripciones, ...), ni las comunicaciones que existan
;;           entre los agentes, etc.
;;
;;           Básicamente un agente que participe en una partida
;;           puede ser de dos tipos:
;;
;;              - Agente activo: Normalmente jugadores, se les
;;                permite o se les inquiere a efecuar acciones
;;                que puedan cambiar el estado del juego
;;
;;              - Agentes pasivos: Son agentes que observan la
;;                partida, aunque opcionalmente pueden atender
;;                alguna petición de otros agentes o el servidor.
;;                Entre ellos estarían agentes para la 
;;                visualización de la partida, agentes para
;;                almacenar un log de la partida y/o recabar
;;                estadísticas o agentes organizadores de
;;                partidas.
;;
;;           Un agente se compone de una lista de propiedades que
;;           habrán de pasarse a la función PET-INI cuando comience
;;           una partida y que incluye un grupo de funciones. Entre
;;           este grupo de propiedades pueden encontrarse las 
;;           desritas en PT-BASE:KEYWORDS-FUN-AG.
;;
;;           Los símbolos accesibles más importantes son los
;;           incluidos en "envio-fun-package.lisp". Ver la
;;           documentación existente para cada una de ellos.
;;
;;           Se intenta que este package sea independiente del
;;           juego que sirva el controlador del juego. Por ello
;;           las referencias/dependencias al juego PlanTron se
;;           iran tratando de eliminar.
;;
;;           Para utilizar una interface no funcional ha de
;;           reescribirse el código, cambiando básicamente las
;;           primitivas funcall y apply por otras equivalentes
;;           al sistema de comunicación a emplear.
;;
;;
;; Dependencias:
;;
;;           Hace uso de funciones del paquete PT-BASE que 
;;           lleva el control del juego.
;;
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (13/04/2006): Primera versión. Extraida
;;             de tron.0.1.lsp. Vease tron-base.lsp
;;
;;           - Ver 0.2 (04/05/2006): Cambio envios a funciones
;;             de finalización, inclusión de identificador de
;;             jugador para agentes activos o nil para agentes
;;             pasivos. Inclusión de funciones envio-evento*.
;;             En envio-ini se envía listas con sublistas de
;;             strings agentes - descripción de agentes
;;             jugadores, y no sólo símbolos.
;;
;;           - Ver 0.2.1 (11/05/2006): Mejorada documentación.
;;             cambio nombre función envio-camb-est. Cambio
;;             nombre variables lista de funciones.
;;
;;           - Ver 0.3 (12/05/2006): Incluida funciones para
;;             el seguimiento de consultas en envio-pet-ini.
;;             Modificaciones de funciones consulta-* para 
;;             permitir seguimiento.
;;
;;           - Ver 0.3.1 (30/05/2006): Incluida función 
;;             consulta-turno y consulta-contador.
;;
;;           - Ver 0.4 (21/4/2007): Cambio de nombre de tron-comm a 
;;             envio. Cambio extensión de "lsp" a "lisp". Cambio 
;;             también nombre de variables, funciones, etc de "tron" a 
;;             "pt".
;;
;;           - Ver 1.0 (1/05/2007): Renovación profunda de esta
;;             interface de comunicaciones. Cambio de nombre del
;;             package de "PT-COMM" a "ENVIO". Cambio de nombre
;;             de muchas funciones, modos de funcionamiento, 
;;             almacenamientos, etc.
;;
;;           - Ver 1.0.1 (8/05/2007): Retoques en la documentación
;;
;;           - Ver 1.0.2 (19/05/2007): Eliminado contador y trasladado
;;             a agente de utilidades
;;
;;           - Ver 1.1 (24/05/2007): Añadido posiblidad de keys
;;             de agentes no preestablecidas para prestar servicios
;;             con keywords-nofun-ag y broadcast
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - Escoger elementos a exportar
;;
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Precisar param-juego y resultado-juego
;;           - ¿Pasar a servidor juego lista deresultados de cada
;;             llamada dentro de envio-evento*?
;;           - Seleeccionar valor a devolver por pet-ini
;;           - ¿En vez de empezar con dim-tab-def fija obtener tamaño
;;             desde la terminal?
;;           - Utilizar interface de comunicaciones vía sockets o
;;             similar (irc, http, kqml, pvm, ...)
;;           - Añadir un gestor más generico (¿otro paquete?) que
;;             permita llamadas más simples a pet-ini-tron
;;           - Intentar independizar de juego
;;           - ¿Separar interface para servidor de interface agentes?
;;             (seguridad)
;;           - ¿Definición de my-warning está bien aquí?
;;           - Hacer que consultas con parámetro jugador acepte, además
;;             string con nombre de jugador.
;;           - Que se les pase a las funciones de seguimientos de
;;             consultas el que solicitó la consulta.
;;           - ¿Mensajes no boradcast?
;;
;;
;;           - Mejoras en estilo de programación
;;             ---------------------------------
;;           - Unificar/armonizar interface de llamadas de agentes
;;           - ¿llamadas con pt-base: es lo mejor? (dependen juego)
;;           - En actualizar-lis-fun-ag iterar sobre las keys del
;;             agente, no sobre pt-base:keywords-fun-ag, para asi
;;             detectar keys erroneas
;;


(unless (find-package "ENVIO")
  (load "envio-fun-package"))

(in-package "ENVIO")


;;
;; Variables versión
;; -----------------

(defconstant version "fun-1.1" "Versión del package ENVIO.")

(defun version ()
  "Devuelve la versión del paquete ENVIO que se ha cargado."
  version)


;;
;; Importamos-exportamos símbolos de pt-base 
;; -----------------------------------------

;; para tipos de jugadores
(shadowing-import pt-base:tipos-jug)
(export pt-base:tipos-jug)

;; para tipos de agentes
(shadowing-import pt-base:tipos-ag)
(export pt-base:tipos-ag)

;; para características de agentes
;(shadowing-import pt-base:keywords-fun-ag)
;(export pt-base:keywords-fun-ag)


;;
;; Constantes y variables globales
;; -------------------------------

(defconstant keywords-fun-ag pt-base:keywords-fun-ag
  "Lista de posibles keywords de los distintos tipos de funciones que pueden tener los agentes para comunicarse. Ver PT-BASE:KEYWORDS-FUN-AG.")

(defconstant keywords-nofun-ag pt-base:keywords-nofun-ag
  "Lista de keywords que no son de tipos de funciones de agentes. Ver PT-BASE:KEYWORDS-NOFUN-AG. ")

(defvar lis-fun-ag nil
  "Lista asociativa con claves los distintos tipos de funciones (ver PT-BASE:KEYWORDS-FUN-AG) y datos las listas las distintas funciones correspondientes a los agentes jugadores existentes.")

(defvar lis-key-ag nil
  "Array para almacenar las lista de keys (ver PT-BASE:KEYWORDS-FUN-AG) de cada agente")

(defvar turno-jugador nil
  "Indica quien es el jugador que en estos momentos esta en posesión del turno, o NIL si ninguno.")

(defvar num-jug nil
  "Indica el número de jugadores (total de agentes) de la partida actual")


;;
;; Warnings
;; --------

(defun my-warning (datum &rest arguments)
  "Enviar un warning a la salida de error."
  (apply #'format 
	 `(,*error-output* ,(format nil "Warning: ~A~%" datum) ,@arguments)))


;;
;; Funciones para creación de listas de listas de keywords
;; -------------------------------------------------------

(defmacro fun-id (key)
  "Indica si KEY se trata de la key de una función identificada."
  `(member ,key pt-base:keywords-fun-id-ag))

(defun obt-lis-elm (key fun id-ag)
  "Obtiene lista de elementos a instertar posteriormente en lis-fun-ag."
  (mapcar #'(lambda (fun)
	      (if (fun-id key)
                  ; Si es función identificada añadir id-ag 
		  (list id-ag fun)
		fun))
	  (if (consp fun)
	      fun
	    (list fun))))

(defun inserta-key-dato (key dato lis-fun-ag)
  "Inserta KEY y DATO en LIS-FUN-AG."
  (let ((lis-key (assoc key lis-fun-ag)))
    (if lis-key
	; Ya la key está en lis-fun-ag, insertar fun
	(setf (rest lis-key) (cons dato (rest lis-key)))
      ; La key no está en lis-fun-ag, crearla e insertar
      (setf lis-fun-ag (acons key (list dato) lis-fun-ag))))
  ; Devolver siempre lis-fun-ag
  lis-fun-ag)

(defun actualizar-key-lis-fun-ag (key fun id-ag lis-fun-ag)
  "Actualiza la lista de keywords LIS-FUN-AG con la función FUN perteneciente al agente con identificador ID-AG en la key KEY si FUN no es NIL."
  (dolist (elm (obt-lis-elm key fun id-ag) lis-fun-ag)
    (setf lis-fun-ag (inserta-key-dato key elm lis-fun-ag))))

;(defun actualizar-lis-fun-ag (ag id-ag lis-fun-ag)
;  "Actualiza la lista de keywords LIS-FUN-AG con las funciones incluidas en el agente con nombre AG."
;  (dolist (key pt-base:keywords-fun-ag lis-fun-ag)
;    (let ((fun (getf ag key)))
;      (if fun        ; Hay función
;	  (setf lis-fun-ag (actualizar-key-lis-fun-ag  key
;						       fun
;						       id-ag
;						       lis-fun-ag))))))

(defun actualizar-lis-fun-ag (ag id-ag lis-fun-ag)
  "Actualiza la lista de keywords LIS-FUN-AG con las funciones incluidas en el agente con nombre AG."
  (do* ((elmag ag (rest (rest elmag)))
	(key (first elmag) (first elmag))
	(fun (second elmag) (second elmag)))
      ((endp elmag) lis-fun-ag)
    (if (not (member key keywords-nofun-ag))     ; Es función de agente
	(setf lis-fun-ag (actualizar-key-lis-fun-ag  key
						     fun
						     id-ag
						     lis-fun-ag)))))


;;
;; Funciones de apoyo para comunicación
;; ------------------------------------

(defun envio-evento-fun (fun &rest lis-datos)
  "Enviar evento mediante función FUN con los datos LIS-DATOS."
  (apply fun lis-datos))

(defun envio-evento-ident (lis-fun-rec lis-datos)
  "Enviar evento y datos LIS-DATOS las funciones determinadas en LIS-FUN-REC, indicándoles tambíen su identificador de jugador."
  (dolist (jug-fun lis-fun-rec t)
    (let ((jug (first jug-fun))
	  (fun (second jug-fun)))
      (setf turno-jugador jug)
      (if (null (apply #'envio-evento-fun fun jug lis-datos))
	    (my-warning "Jugador ~S no acepta evento." jug))))
  (setf turno-jugador nil)
  t)


(defun envio-evento-sin-ident (lis-fun-rec lis-datos)
  "Enviar evento no identificado con datos LIS-DATOS las funciones en LIS-FUN-REC."
  (dolist (fun lis-fun-rec t)
    (apply #'envio-evento-fun fun lis-datos)))


;;
;; Interface de consultas servidor de juego pt-base ==> agentes
;; ------------------------------------------------------------

(defun broadcast (tipo-fun &rest lis-datos)
  "Enviar evento de tipo TIPO-FUN con la lista de datos LIS-DATOS a todos los agentes con funciones TIPO-FUN (evento broadcast)."
  (let ((lis-fun-tipo-fun (rest (assoc tipo-fun lis-fun-ag))))
    (if lis-fun-tipo-fun
	(if (fun-id tipo-fun)
	    (envio-evento-ident lis-fun-tipo-fun lis-datos)
	  (envio-evento-sin-ident lis-fun-tipo-fun lis-datos))
      t)))

(defun puntual (jugador tipo-fun &rest lis-datos)
  "Enviar evento de tipo TIPO-FUN con la lista de datos LIS-DATOS al agente JUGADOR con su función TIPO-FUN, si existe dicha función (evento puntual)."
  (setf turno-jugador jugador)
  (prog1
      (if (fun-id tipo-fun)
	  (apply #'envio-evento-fun
		 (getf (aref lis-key-ag jugador) tipo-fun)
		 (cons jugador lis-datos))
	(apply #'envio-evento-fun
	       (getf (aref lis-key-ag jugador) tipo-fun)
	       lis-datos))
    (setf turno-jugador nil)))


;;
;; Interface de consultas agentes ==> servidor de juego pt-base
;; ------------------------------------------------------------

(defun pet-ini (lis-lis-key-ag tipo-juego)
  "Lanza una partida entre los agentes descritos por listas de propiedades en LIS-LIS-KEY-AG (ver PT-BASE:KEYWORDS-FUN-AG). TIPO-JUEGO define el tipo de juego que se pretende jugar en la partida (ver servidor de juego PT-BASE)."
  ; Inicializamos num-jug
  (setf num-jug (length lis-lis-key-ag))
  (let ((id-ag 0))
    ; Generamos lista y array de funciones de agentes
    (setf lis-key-ag (make-array (1+ num-jug) :initial-element nil))
    (setf lis-fun-ag nil)
    (dolist (ag lis-lis-key-ag)
      (setf id-ag (1+ id-ag))
      (setf (aref lis-key-ag id-ag) ag)
      (setf lis-fun-ag (actualizar-lis-fun-ag ag id-ag lis-fun-ag)))

    ; Llamada a servidor pt
    (or (pt-base:inicio tipo-juego)          ; Tipo de juego
	(error "Función envio-pet-ini falla"))))

(defun pet-consulta-accion ()
  "Solicita al servidor del juego que le sea enviado una CONSULTA ACCIÓN al jugador que tiene el turno en este momento. Usado para juegos asíncronos."
  (pt-base:get-accion turno-jugador))

(defun pet-gen-planta (posicion &optional inc-ene)
  "Solicita al servidor del juego la colocación de una planta del jugador (generador de plantas) que tiene el turno actual en POSICION. Opcionalmente INC-ENE supone el incremento de energía que adquirirá el jugador generador de la planta."
  (pt-base:gen-planta turno-jugador posicion inc-ene))

(defun pet-transf-ene (jug-des &optional catidad)
  "Solicita al servidor del juego la transferencia al jugador JUG-DES de CANTIDAD unidades de energía"
  (pt-base:transf-ene turno-jugador jug-des catidad))

(defun consulta-leyenda ()
  "Solicita al servidor del juego la leyenda que define el juego actual implementado."
  (pt-base:leyenda))

(defun consulta-pos-jug (jugador)
  "Posición actual del jugador con identificador JUGADOR. El resultado depende de lo enviado por el servidor de juegos (ver PT-BASE:POS-JUGADOR)."
  (broadcast :seg-cposj turno-jugador jugador)
  (pt-base:pos-jugador jugador))

(defun consulta-est-jug (jugador)
  "Estado actual del jugador con identificador JUGADOR. El resultado depende de lo enviado por el servidor de juegos (ver PT-BASE:EST-JUGADOR)."
  (broadcast :seg-cestj turno-jugador jugador)
  (pt-base:est-jugador jugador))

(defun consulta-pos-tab (pos &optional dir)
  "Contenido actual de una posición del tablero POS más opcionalmente desplazamiento en la dirección DIR sobre ella. El resultado depende de lo enviado por el servidor de juegos (ver PT-BASE:POS-TABLERO)."
  (broadcast :seg-cpost turno-jugador pos dir)
  (pt-base:pos-tablero pos dir))

(defun consulta-tipo (id)
  "Tipo actual de el valor ID, que puede ser el contenido del tablero o bien el identificador de un jugador. Los tipos posibles a devolver estan definidos por el servidor de juegos (ver PT-BASE:TIPOS-JUG)."
   (pt-base:tipo-jug id))

(defun consulta-tab ()
  "Solicita la matriz del tablero del juego."
  (pt-base:tablero))


;;
;; Interface de consultas agentes ==> servidor de comunicaciones
;; -------------------------------------------------------------

(defun consulta-turno ()
  "Solicita el identificador que en estos momentos esta en posesión del turno de juego."
  turno-jugador)

(defun consulta-lis-key-jug (id)
  "Solicita la lista de keywords del agente con identificador ID."
  (if (and (numberp id) (plusp id) (<= id num-jug))
      (aref lis-key-ag id)))


;;
;; Interface de consultas agentes ==> agentes
;; ------------------------------------------

(defun mensaje (mensaje)
  "Envío de un mensaje MENSAJE a todos los agentes que puedan recibir mensajes, por ejemplo los agentes de PT-VIS."
  (broadcast :rec-msj turno-jugador mensaje))

