;; Programa: Package con agentes jugadores generadores de
;;           plantas de PlanTron
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    25/04/2007
;; Comentarios:
;;
;;           Los agentes definidos utilizan el entorno de juego
;;           PlanTron, definido principalmente por los packages
;;           PT, PT-BASE Y ENVIO.
;;
;;           Las funciones accesibles más importantes son:
;;
;;           ....
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;; Agradecimientos:
;;           En gran parte este agente ha sido inspiradon en el 
;;           entorno y código de "GA applet" de David Eck:
;;           http://math.hws.edu/xJava/GA/
;;
;; Dependencias:
;;
;;           Hace uso del paquete que contiene la interface (ENVIO)
;;           con el control del juego (PT-BASE).
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (25/04/2007): Primera versión.
;;
;;           - Ver 0.2 (1/05/2007): Ajustes para versión con nueva
;;             interface de comunicaciones en package "ENVIO"
;;
;;           - Ver 0.2.1 (2/05/2007): Reparado error en :tipo-ag
;;             que impedía contar agentes gen-pla. Ajuste de 
;;             *ratio-plantas*
;;
;;           - Ver 0.2.2 (26/05/2007): Incluida posibilidad o no
;;             de renacer las plantas con *renacer*
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - En ocasiones gp-cumulo pone una planta sobre rastro con
;;             *plantar-sobre-rastro* a nil 
;;           - Que se pueda llamar con varios generadores de plantas
;;             del mismo o distinto tipo (conflictos fuentes, aleat, etc.)
;;           - El agente mitad a veces deja algunas columnas sin explorar
;;             cuando encuentra un cúmulo de plantas
;;           - En los cálculos del número de plantas inicial mínimo
;;             necesario para poder llenar todo el espacio no se tuvo 
;;             en cuenta la creación de nuevas plantas
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Afinar valores por defecto de variables: sin mensajes
;;             de fallos pero rápidos
;;           - Añadir más tipos de agentes generadores
;;           - Que las plantas mueran por si solas tras un tiempo de
;;             vida no fijo
;;           - Que las plantas se creen no (solo) cuando mueren
;;           - Generadores más simples, que no creen nuevas plantas,
;;             solo creen plantas iniciales
;;           - Que los agentes generadores sean reentrantes,esto es
;;             se puedan llamar al mismo agente como varios jugadores
;;             con variables de comportamiento distintas y no haya 
;;             conflictos con ellas
;;
;;           - Mejoras en estilo de programación
;;             ---------------------------------
;;           - Que haya una variable random por cada generador o id
;;           - Implementar Fase 2 de busca-libre-cerca
;;
;;


;;
;; Definición del package
;; ----------------------

(defpackage "PT-AGENTES-GENPLA"
  (:nicknames "PT-A-GP")
  (:use "COMMON-LISP")
  (:export "AGENTES" "VERSION" "AGENTE-GP-ALEAT" "AGENTE-GP-ALEAT-INI" "AGENTE-GP-CUMULOS" "AGENTE-GP-CUMULOS-INI" "*RATIO-PLANTAS*" "*RATIO-MAX-INTENTOS-BPL*" "*ENERGIA-PLA*" "*RENACER*" "*PLANTAR-SOBRE-RASTRO*" "*RATIO-FUENTES-PLANTAS*" "*RATIO-MAX-INTENTOS-BLC*" "*RATIO-FUENTES*" "*PROB-MOD-FUENTE*"))


(in-package "PT-AGENTES-GENPLA")



;;
;; Constante versión
;; -----------------

(defconstant version "0.2.2"
  "Versión del package PT-AGENTES-GENPLA.")


;;
;; Constante con lista de agentes presentes en este package
;; --------------------------------------------------------

(defconstant agentes
  '((:es-comido agente-gp-aleat
		:inicio agente-gp-aleat-ini
		:tipo-ag envio:gen-pla )
    (:es-comido agente-gp-cumulos
		:inicio agente-gp-cumulos-ini
		:tipo-ag envio:gen-pla ))
  "Agentes tipo GEN-PLA para jugar a PlanTron.")


;;
;; Variables
;; ---------

(defvar *ratio-plantas* 0.025
  "Ratio de número de plantas por unidad de superficie")

(defvar *ratio-max-intentos-bpl* 0.06
  "Ratio del máximo número de intentos aleatorios de buscar posición libre para crear planta")

(defvar *energia-pla* 10
  "Energía por planta. Al crear una planta el generador gana esa energía, cuando muere una planta el generador pierde esa energía, transfiriéndola al comedor")

(defvar *plantar-sobre-rastro* t
  "Indica si es posible plantar una planta sobre el rastro de un jugador o no.")

(defvar *renacer* t
  "Indica si al morir una planta renace nuevamente en otro lado.")

(defvar dim-tab nil
  "Dimensiones del tablero")

(defvar max-intentos-bpl nil
  "Máximo número de intentos aleatorios de buscar posición libre para crear planta") 

(defvar random-state-gp (make-random-state *random-state*)
  "Estado random para el agente-gp-aleat.")


;;
;; Funciones de apoyo
;; ------------------

(defun posible-plantar-p (tipo-cont)
  "Mira si es posible plantar una planta dado el contenido de tipo TIP-CONT de una casilla."
  (or (null tipo-cont) 
      (and *plantar-sobre-rastro* 
	   (eq tipo-cont 'envio:jugador))))

(defun inc-pos (pos dir)
  "Incrementa la posición POS en dirección DIR."
  (case dir
    (0 (list (first pos) (1- (second pos))))
    (1 (list (1+ (first pos)) (second pos)))
    (2 (list (first pos) (1+ (second pos))))
    (3 (list (1- (first pos)) (second pos)))))

(defun dir-inc (dir dir-ini)
  "Calcula la dirección DIR relativa a DIR-INI, esto es, gira DIR-INI pasos en sentido agujas del reloj sobre DIR."
  (rem (+ dir dir-ini) 4))

(defun elige1-y-limpia (lista)  ; No usado de momento
  "Elige aleatoriamente un elemento de la lista LISTA y devuelve dicho elemento y una nueva lista igual a LISTA pero sin el elemento"
  (let ((elm (nth (random (length lista) random-state-gp)
		  lista)))
    (values elm (remove elm lista :count 1))))

(defmacro return-if-not-nil (expr)
  "Ejecuta RETURN devolviendo EXPR solo si EXPR es distinto de NIL, sino devuelve NIL"
  `(let ((res-expr ,expr))
     (if res-expr
         (return res-expr))))

(defun gen-pos-aleat-libre (&optional (max-intentos max-intentos-bpl))
  "Busca una posición aleatoria libre"
  (let (posxy)
    (dotimes (i max-intentos nil)
      (setf posxy (list (random (first dim-tab) random-state-gp)
			(random (second dim-tab) random-state-gp)))
      (if (posible-plantar-p (envio:consulta-tipo (envio:consulta-pos-tab posxy)))
	  (return posxy)))))

(defun gen-planta (id posxy)
  "Genera una nueva planta en POSXY"
  (if posxy
      (if (envio:pet-gen-planta posxy *energia-pla*)
	  t
	(envio:my-warning "Agente ~S se le denegó generar nueva planta" id))
    (envio:mensaje "No encuentro hueco para generar nueva planta")))

;;
;; Agente generador de plantas aleatorias
;; --------------------------------------

(defun agente-gp-aleat-ini (id tipo-juego)
  "Inicializador de agente jugador PlanTron tipo genpla que genera plantas aleatoriamente dentro del tablero."
  ; Extraemos dimensiones del tablero
  (setf dim-tab (getf tipo-juego :dim-tab))
  ; Calculamos número máximo de intentos aleatorios
  (setf max-intentos-bpl (round (* *ratio-max-intentos-bpl* (first dim-tab) (second dim-tab))))
  ; Generación de plantas
  (dotimes (i (round (* *ratio-plantas* (first dim-tab) (second dim-tab))) t)
    (gen-planta id (gen-pos-aleat-libre max-intentos-bpl))))


(defun agente-gp-aleat (id-gpla id-gdir posicion)
  "Agente jugador PlanTron tipo genpla que genera planta aleatoriamente dentro del tablero."
  (declare (ignore posicion))
  ; Transferir energia al comedor
  (envio:pet-transf-ene id-gdir *energia-pla*)
  (if *renacer*
    ; Cuando se coman una planta generar una nueva
    (gen-planta id-gpla (gen-pos-aleat-libre max-intentos-bpl))))


;;
;; Variables para agente generador de plantas en cumulos
;; -----------------------------------------------------

(defvar *ratio-fuentes-plantas* 0.1
  "Ratio de número de fuentes sobre el ratio de número de plantas por unidad de superficie")

(defvar *ratio-max-intentos-blc* 1  ; FIXME disminuir con impl. fase 2 blc
  "Ratio del máximo número de intentos de buscar libre cerca para crear planta")

(defvar *ratio-fuentes* (/ *ratio-plantas* 20)
  "Ratio de número de fuentes por unidad de superficie")

(defvar *prob-mod-fuente* 0.05
  "Probabilidad de modificar fuente")

(defvar num-fuentes nil
  "Número de fuentes a utilizar")

(defvar fuentes nil
 "Array para almacenar las fuentes")

(defvar max-intentos-blc nil
  "Máximo de intentos en buscar-libre-cerca")

(defvar cont-int-blc nil
  "Contador de intentos en buscar-libre-cerca")

;;
;; Funciones para agente generador de plantas en cumulos
;; -----------------------------------------------------

(defun busca-prof-libre (posxy)
  "Busca en profundidad una casilla libre cerca de POSXY, eligiendo aleatoriamente la posición en que comenzar y limitando el número de intentos"
  (setf cont-int-blc (1- cont-int-blc))
  (let ((tipo-cont (envio:consulta-tipo (envio:consulta-pos-tab posxy))))
    (cond
     ((posible-plantar-p tipo-cont) posxy)    ; Si es posible, plantar ahí
     ((minusp cont-int-blc) nil)              ; Si fin contador, fallar
     ((eq tipo-cont 'envio:obstaculo) nil)  ; Si sobre obstáculo, fallar
     ;((eq tipo-cont 'envio:planta)   ; Si sobre planta
     (t                                       ; Si sobre planta o jugador:
      ; Busca en las 4 direcciones empezando en una aleatoria
      (let ((dir-ini (random 4 random-state-gp)))
	(dotimes (i 4 nil)
	  (if (minusp cont-int-blc)
	      (return nil)
	    (return-if-not-nil
	     (busca-prof-libre (inc-pos posxy (dir-inc i dir-ini)))))))))))
     ;(t nil))))                           ; Si choca con algo, fallar


(defun busca-libre-cerca (posxy &optional (max-intentos max-intentos-blc))
  "Busca una casilla libre cerca de POSXY"
  ; Dos fases:
  ; Fase 1: Se sale con búsqueda en profundidad pisando plantas y
  ; jugadores, pero no pisando obstaculos
  (setf cont-int-blc max-intentos)
  (busca-prof-libre posxy))
  ; Fase 2: Avanza en línea recta en dirección aleatoria hasta libre

(defun gen-una-fuente (id id-fuente &optional (max-intentos max-intentos-bpl))
  "Genera la fuente aleatoria ID-FUENTE para generar plantas"
  (declare (ignore id))
  (let ((posxy (gen-pos-aleat-libre max-intentos)))
    (if posxy
	(setf (aref fuentes id-fuente)
	      posxy)
      (envio:mensaje "No puedo generar nuevo cúmulo"))))

(defun gen-fuentes (id &optional (max-intentos max-intentos-bpl))
  "Genera las fuentes para generar plantas"
  (setf num-fuentes (round (* *ratio-fuentes-plantas*
			      *ratio-plantas* 
			      (first dim-tab) (second dim-tab))))
  (setf fuentes (make-array num-fuentes :initial-element nil))
  (dotimes (i num-fuentes t)
    (gen-una-fuente id i max-intentos)))

(defun gen-planta-cumulo (id
			  &optional 
			  (max-intentos-1 max-intentos-blc)
			  (max-intentos-2 max-intentos-bpl))
  "Genera planta en cumulo en base a una fuente elegida aleatoriamente"
  (let ((id-fuente (random num-fuentes random-state-gp)))
    ; Genera planta según fuente elegida
    (gen-planta id
		(busca-libre-cerca (aref fuentes 
					 id-fuente)
				   max-intentos-1))
    ; Ver si modificar fuente
    (if (< (random 1.0 random-state-gp) *prob-mod-fuente*)
	; Plantas comienzan a crecer en otro lado
	(gen-una-fuente id id-fuente max-intentos-2))))


;;
;; Agente generador de plantas en cumulos
;; --------------------------------------

(defun agente-gp-cumulos-ini (id tipo-juego)
  "Inicializador de agente jugador PlanTron tipo genpla que genera plantas en cúmulos dentro del tablero."
  ;Extraemos dimensiones del tablero
  (setf dim-tab (getf tipo-juego :dim-tab))
  ; Calculamos número máximo de intentos
  (setf max-intentos-bpl (round (* *ratio-max-intentos-bpl* (first dim-tab) (second dim-tab))))
  (setf max-intentos-blc (round (* *ratio-max-intentos-blc* (first dim-tab) (second dim-tab))))
  ; Generación de fuentes
  (gen-fuentes id max-intentos-bpl)
  ; Generación de plantas
  (dotimes (i (round (* *ratio-plantas* (first dim-tab) (second dim-tab))) t)
    (gen-planta-cumulo id)))


(defun agente-gp-cumulos (id-gpla id-gdir posicion)
  "Agente jugador PlanTron tipo genpla que genera planta en cúmulos dentro del tablero."
  (declare (ignore posicion))
  ; Transferir energia al comedor
  (envio:pet-transf-ene id-gdir *energia-pla*)
  (if *renacer*
    ; Cuando se coman una planta generar una nueva
    ; Seleccionar una fuente aleatoriamente y buscar cerca,
    ; plantar y ocasionalmente cambiar fuente
    (gen-planta-cumulo id-gpla)))
