;; Programa: Package para agentes visualizadores del juego PlanTron
;;
;; Asignatura: Introducción a la Inteligencia Artificial
;;             ETSII - ULL
;; Práctica: (sistemas de producción)
;; Profesor: Patricio García (patricio AT etsii.ull.es)
;; Fecha:    12/04/2006
;; Comentarios:
;;
;;           Las funciones accesibles más importantes son:
;;
;;           Ver la documentación existente para cada una de ellas.
;;
;;           Para su utilización es aconsejable usar el cargador
;;           del entorno de juego PlanTron: "pt-loader.lisp".
;;
;;
;; Dependencias:
;;           Hace uso del paquete que contiene la interface (envio)
;;           con el control del juego (pt-base).
;;
;; Modificaciones:
;;
;;           - Ver 0.1 (13/04/2006): Primera versión. Extraida
;;             de tron.0.1.lsp. Véase tron-base.lsp y tron-comm*.lsp.
;;             Realizada ocultación del cursor. Incluido tiempo de
;;             espera al final. Arreglado visulaización de
;;             descripción de funciones agentes y nombres.
;;
;;           - Ver 0.1.1 (03/05/2006): Reparado bug que desplazaba
;;             bordes del tablero.
;;
;;           - Ver 0.1.2 (03/05/2006): Cambio en fin-vis-tron debido
;;             a cambio en tron-comm de ver "fun-0.2".
;;
;;           - Ver 0.2 (05/05/2006): Inclusión de variable
;;             *tron-vis-stream* que permite visualizar el tablero
;;             en otro streamer, ej (open "/dev/pts/12" :direction 
;;             :output :if-exists :overwrite). Buscarla con "
;;             w $LOGNAME | grep "w $LOGNAME"". Añadido versión
;;             del package en el borde superior izquierdo.
;;
;;           - Ver 0.3 (12/05/2006): Incluida funciones para
;;             el seguimiento de consultas, debido a cambios
;;             en tron-comm de ver "fun-0.3". Generados varios
;;             visualizadores según seguimientos. Eliminado bug
;;             que no limpiaba bien la línea de estado
;;             (con "/dev/pts/...").
;;
;;           - Ver 0.3.1 (20/05/2006): Corregido bug en 
;;             borra-pinta-seg que impedía pintar ",". Cambiado
;;             seguimiento para que pinte en cuanto se recibe.
;;
;;           - Ver 0.3.2 (27/05/2006): Añadido envio-punteo-pos.
;;             Añadido envio-mensaje.
;;
;;           - Ver 0.4 (21/4/2007): Cambio de nombre de tron-vis a 
;;             pt-vis. Cambio extensión de "lsp" a "lisp". Cambio 
;;             también nombre de variables, funciones, etc de "tron"
;;             a "pt".
;;
;;           - Ver 0.5 (1/05/2007): Ajustes para versión con nueva
;;             interface de comunicaciones en package "ENVIO"
;;
;;           - Ver 0.5.1 (2/05/2007): Arreglado: problemas con scroll
;;             en barra de estado sobrepasada, errores con 
;;             visualizadores 4 ó 5, evitado pintar seguimientos
;;             hechos por el propio pt-vis. Incluido visualizar
;;             nombres de agentes pasivos al principio. Reparado
;;             error en :tipo-ag que impedía contar agentes pasivos
;;
;;           - Ver 0.5.2 (23/05/2007): Añadido seguimiento de :seg-npla.
;;             Arreglado problemas al visualizar seguimientos de
;;             consultas fuera del tablero. Añadido funcionalidad *step*
;;
;;           - Ver 0.5.3 (24/05/2007): Cambios debidos a keywords para
;;             servicios :refresco y :punteo
;;
;;
;; Por hacer:
;;           - Corrección de Bugs
;;             ------------------
;;           - interpreted-function ha de ser eval:interpreted-function,
;;             pero esto provoca un error en el gcl de paquete 
;;             no encontrado.
;;           - Ver como conjuntar envio-refresco-tab en interface de comm.
;;           - A veces aparece (en CMUCL): ** Closed file "/dev/pts/14"
;;           - No se visualizan los caracteres de seguimiento en sbcl
;;
;;
;;           - Mejoras en funcionalidades
;;             --------------------------
;;           - Verificar funcionamiento para <> de 2 jugadores.
;;             Problemas caracteres, colores, salto línea status.
;;           - Tratar de independizar más la visualización.
;;           - Revisar nombre de variables.
;;           - Agregar la documentación a cada función importante.
;;           - ¿Que vis-tron y vis-jug-tron impriman ! en el mismo color?
;;           - ¿Añadir aquí, o en agente-humano, un agente que permita
;;             "pausar" la ejecución?
;;           - Hacer versión con gcl-tk.
;;           - Revisar tipos de visualizadores.
;;           - Que los mensajes recibidos se puedan imprimir opcinalmente
;;             también en una consola distinta.
;;           ~ Hacer pasar las funciones de interface de consultas
;;             agentes-jugadores ==> agente visualizador por envio
;;           - Imprimir al empezar también versiones de agentes
;;           - Que puedan recibir mensajes también de agentes pasivos
;;           + Variable para paso a paso *step*
;;           - Que el agente visualizador 0 pinte al inicio también
;;             las plantas
;;           - Que *step* valga tambien para visualizador 1
;;           - Añadir seguimientos de :seg-mpla
;;
;;




;;
;; Definición del package
;; ----------------------

(defpackage "PT-VIS"
  (:nicknames "PT-VIS-XTERM" "PT-V" "PT-V-X")
  (:use "COMMON-LISP")
  (:documentation "Agentes visualizadores de partidas PlanTron")
  (:export "AGENTES" "VERSION" "VIS-BAJO" "VIS-NORMAL" "VIS-CPOST" "VIS-CDIRJ" "VIS-CPOSJ" "VIS-FULL" "*PT-VIS-STREAM*" "*STEP*" "INI-VIS" "FIN-VIS" "FIN-VIS0" "SEG-POS-VIS" "SEG-DIR-VIS" "SEG-CPOSJ-VIS" "SEG-CDIRJ-VIS" "SEG-CPOST-VIS" "ENVIO-REFRESCO-TAB" "ENVIO-PUNTEO-LIS" "REC-MENSAJE"))

(in-package "PT-VIS")


;;
;; Variables version
;; ------------------

(defconstant version "xterm-0.5.3" "Versión del package PT-VIS.")

(defun version ()
  "Devuelve la versión del PT-VIS que se ha cargado."
  version)


;;
;; Constante con lista de agentes presentes en este package
;; --------------------------------------------------------

(defvar agentes '((:name "VIS-BAJO"
	           :documentation "Visualizador más simple, solo muestra situación al inicio y al final de la partida."
		   :tipo-ag envio:pasivo
		   :inicio ini-vis
		   :refresco envio-refresco-tab
		   :fin fin-vis0)
		  (:name "VIS-NORMAL"
		   :documentation "Visualizador por defecto, no realiza seguimientos de las consultas de los agentes."
		   :tipo-ag envio:pasivo
		   :inicio ini-vis
		   :fin fin-vis
		   :seg-pos seg-pos-vis
		   :seg-npla seg-pos-vis
		   :refresco envio-refresco-tab
		   :punteo envio-punteo-lis
		   :rec-msj rec-mensaje)
		  (:name "VIS-CPOST"
	           :documentation "Visualizador que realiza seguimiento de las consultas sobre el contenido de posiciones del tablero (envio:consulta-pos-tab): caracter \",\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
		   :tipo-ag envio:pasivo
		   :inicio ini-vis
		   :fin fin-vis
		   :seg-pos seg-pos-vis
		   :seg-npla seg-pos-vis
		   :seg-cpost seg-cpost-vis
		   :refresco envio-refresco-tab
		   :punteo envio-punteo-lis
		   :rec-msj rec-mensaje)
		  (:name "VIS-CDIRJ"
		   :documentation "Visualizador que realiza seguimiento de las consultas sobre la dirección de los agentes, caracter \"?\", y el contenido de posiciones del tablero: caracter \",\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
		   :tipo-ag envio:pasivo
		   :inicio ini-vis
		   :fin fin-vis
		   :seg-pos seg-pos-vis
		   :seg-npla seg-pos-vis
		   :seg-cestj seg-cdirj-vis
		   :seg-cpost seg-cpost-vis
		   :refresco envio-refresco-tab
		   :punteo envio-punteo-lis
		   :rec-msj rec-mensaje)
		  (:name "VIS-CPOSJ"
		   :documentation "Visualizador que realiza seguimiento de las consultas sobre la posición de los agentes, caracter \":\", y el contenido de posiciones del tablero: caracter \",\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
		   :tipo-ag envio:pasivo
		   :inicio ini-vis
		   :fin fin-vis
		   :seg-pos seg-pos-vis
		   :seg-npla seg-pos-vis
		   :seg-cposj seg-cposj-vis
		   :seg-cpost seg-cpost-vis
		   :refresco envio-refresco-tab
		   :punteo envio-punteo-lis
		   :rec-msj rec-mensaje)
		  (:name "VIS-FULL"
		   :documentation "Visualizador que realiza todos los seguimientos: consultas sobre la posición de los agentes, caracter \":\", la dirección de los agentes, caracter \"?\", y el contenido de posiciones del tablero: caracter \",\" para consultas absolutas o \"^\", \">\" \"v\" y \"<\" para consultas relativas."
		   :tipo-ag envio:pasivo
		   :inicio ini-vis
		   :fin fin-vis
		   :seg-pos seg-pos-vis
		   :seg-npla seg-pos-vis
		   :seg-cposj seg-cposj-vis
		   :seg-cestj seg-cdirj-vis
		   :seg-cpost seg-cpost-vis
		   :refresco envio-refresco-tab
		   :punteo envio-punteo-lis
		   :rec-msj rec-mensaje))
  "Lista de funciones de agente de visualización")


;;
;; Variables globales
;; ------------------

(defvar *pt-vis-stream* *standard-output*
  "Stream de visualización del tablero PlanTron. Puede ser un fichero de terminal (/dev/pts/*).")

(defvar *step* nil
  "Indica si detenerse y esperar la entrada de una tecla despues de imprimir caracteres de seguimientos y antes de borrarlos. Usar junto a variable *PT-VIS-STREAM* u opción :dev-vis de PT:PARTIDA")

(defvar num-jug-gdir nil
  "Número de jugadores tipo serpiente (gendir) en la partida actual.")

(defvar num-jug-act nil
  "Número de jugadores activos totales (gendir y genplant) en la partida actual.")

(defvar num-jug-pas nil
  "Número de agentes pasivos (visualizadores, logs, contadores, estadísticos, etc.) en la partida actual.")

(defvar pos-jug-gdir nil
  "Array para almacenar las posiciones actuales de cada jugador gendir.")

(defun pos-jug (jug)
  "Cálculo local de la posición del jugador JUG."
  (copy-list (aref pos-jug-gdir (1- jug))))

(defvar msg-jug nil
  "Array para almacenar los mensajes recibidos de cada jugador")

(defvar sleep-cs 0
  "Tiempo en centésimas de segundo entre turnos.")

(defvar dim-tab nil
  "Dimensiones del tablero PlanTron.")

(defvar status-pos '(0 23)
  "Posición (X Y) de comienzo de la línea de estado. Debajo del tablero.")

(defun calc-status-pos ()
  "Calculo de posición (X Y) de comienzo de la línea de estado. Debajo del tablero."
  (list 0 (second dim-tab)))

(defvar *int-ref-est-jug* 5
  "Promedio de movimientos por jugador para actualizar barra de estado")

(defvar int-ref-est nil
  "Número de movimientos para actualizar barra de estado")

(defvar cont-ref-est nil
  "Contador de movimientos para actualizar barra de estado")


;;
;; Funciones específicas según implementación lisp
;; -----------------------------------------------

(let ((lit (lisp-implementation-type)))
  (cond
   ((string-equal lit "GNU Common Lisp (GCL)"))
   ((string-equal lit "CMU Common Lisp")
    (set (find-symbol "*GC-VERBOSE*" "EXTENSIONS") nil))
   ((string-equal lit "CLISP"))
   ((string-equal lit "SBCL"))
   (T ; Implementación desconocida
    (envio:my-warning "Implementación ~S no soportada" lit))))


;;
;; Primitivas para acceso a pantalla
;; ---------------------------------

(defconstant escape-char (code-char 27)
  "Escape char.")

(defconstant command-cursor 
  (concatenate 'string (string escape-char) "[")
  "String de envio de comandos al cursor.")

(defconstant command-cursor-color-ini 
  (concatenate 'string (string escape-char) "[")
  "String de envio de incicio de comando de color al cursor.")

(defconstant command-cursor-color-end "m"
  "String de envio de final de comando de color al cursor.")

(defconstant cursor-color-nothing "0")
(defconstant cursor-color-light-purple "1;35")
(defconstant cursor-color-light-blue "1;34")
(defconstant cursor-color-light-green "1;32")
(defconstant cursor-color-light-brown "0;33")

(defconstant command-title-bar-ini
  (concatenate 'string (string escape-char) "]0;")
  "String de envio de incicio de comando al title bar.")

(defconstant command-title-bar-end (code-char 7)
  "Caracter de envio de final de comando al title bar.")


(defun set-cursor-off (&optional (st *standard-output*))
  "Oculta el cursor."
  (princ command-cursor st)
  (princ "?25l" st))

(defun set-cursor-on (&optional (st *standard-output*))
  "Muestra el cursor."
  (princ command-cursor st)
  (princ "?25h" st))

(defun save-window-cursor-position (&optional (st *standard-output*))
  "Salva la posición del cursor."
  (princ command-cursor st) (princ #\s st))

(defun restore-window-cursor-position (&optional (st *standard-output*))
  "Recupera la posición del cursor salvada."
  (princ command-cursor st) (princ #\u st))

(defun set-window-cursor-position (line column &optional (st *standard-output*))
  "Posiciona el cursor."
  (princ command-cursor st)
  (format st "~D;~DH" line column))

(defun set-window-cursor-color (color &optional (st *standard-output*))
  "Establecer color del cursor."
  (princ command-cursor-color-ini st)
  (format st "~A~A" color command-cursor-color-end))

(defun set-title-bar (msg &optional (st *standard-output*))
  "Imprime mensaje MSG en title-bar."
  (princ command-title-bar-ini st)
  (princ msg st)
  (princ command-title-bar-end st)
  nil)

;;
;; Visualización de estructura de problema
;; ---------------------------------------

(defvar info-ag-jug
  "Array con información de los agentes jugadores")

(defconstant char-jug1 #\+
  "Caracter para visualizar el jugador 1")

(defconstant char-jug2 #\*
  "Caracter para visualizar el jugador 2")

(defconstant char-pla1 #\O
  "Caracter para visualizar las plantas tipo 1")

(defconstant char-pla2 #\@
  "Caracter para visualizar las plantas tipo 2")

(defconstant char-obs #\#
  "Caracter para visualizar los obstáculos")

(defconstant char-end #\!
  "Caracter para visualizar el choque final")

(defconstant color-jug1 cursor-color-light-purple
  "Color para visualizar el jugador 1")

(defconstant color-jug2 cursor-color-light-blue
  "Color para visualizar el jugador 2")

(defvar color-pla1 cursor-color-light-green
  "Color para visualizar las plantas tipo 1")

(defvar color-pla2 cursor-color-light-green
  "Color para visualizar las plantas tipo 2")

(defconstant num-char-color 2)

(defvar color-char-jug
  (make-array (list num-char-color 2) :initial-contents
	      `((,char-jug1 ,color-jug1)
		(,char-jug2 ,color-jug2)))
  "Caracteres y colores para visualizar jugadores")

(defvar color-char-pla
  (make-array (list num-char-color 2) :initial-contents
	      `((,char-pla1 ,color-pla1)
		(,char-pla2 ,color-pla2)))
  "Caracteres y colores para visualizar plantas")

(defvar color-obs cursor-color-nothing
  "Color para visualizar los obstáculos")

(defvar color-fondo cursor-color-nothing
  "Color para visualizar el fondo")

(defvar color-end cursor-color-light-brown
  "Color para visualizar el choque final")

(defvar elm-mat-char-color nil
  "Array para transformar el contenido de elemento del tablero a carácter y color")

(defvar lis-con-seg nil
  "Lista de consultas realizadas pendientes de pintar.")

(defvar lis-con-pin nil
  "Lista de consultas realizadas ya pintadas pendientes de borrar.")

(defvar consulta-no-vis t
  "Indica si no es el propio visualizador el que realiza las consulta a ENVIO para que no se pinte.")

(defvar leyenda nil
  "Leyenda a imprimir en title-bar")

(defmacro col-jug (jug)
  `(if ,jug
       (aref elm-mat-char-color ,jug 1)
     color-fondo))

(defmacro car-jug (jug)
  `(if ,jug
       (aref elm-mat-char-color ,jug 0)
     #\Space))

(defun vis-pos-color (posxy color obj &optional (st *standard-output*))
  "Posiciona el cursor en la posición POSXY, si POSXY es no NIL, activa le color COLOR e imprime el objeto OBJ"
  (if posxy
      (set-window-cursor-position (1+ (second posxy)) (1+ (first posxy)) st))
  (set-window-cursor-color color st)
  (princ obj st))


(defun vis-prev (&optional (st *standard-output*))
  "Visualización de información previa a partida PT"
  ; Imprime datos de agentes jugadores activos
  (dotimes (i num-jug-act)
    (format st "~%Jugador ~D: ~A~%" (1+ i) (aref info-ag-jug (1+ i) 0))
    (format st "~A~%" (aref info-ag-jug (1+ i) 1)))

  ; Imprime datos de agentes pasivos
  (dotimes (i num-jug-pas)
    (let* ((id (+ num-jug-act 1 i))
	   (lis-keys (envio:consulta-lis-key-jug id)))
      (format st "~%Agente pasivo (~D): ~A~%" id (getf lis-keys :name))
      (format st "~A~%" (getf lis-keys :documentation))))

  ; Espera para empezar
  (sleep (ceiling (* 12 sleep-cs) 100))

  (format st "~%Comienza la partida ")
  (if (equal *standard-output* *pt-vis-stream*)
      (format st "...~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%~%")
    (format st "en ~S ...~%" *pt-vis-stream*))
  (force-output st))


(defun clear-status (&optional (st *standard-output*))
  "Se posiciona en la línea de estado y la borra, dejando el cursor al principio de ella."
  (vis-pos-color status-pos cursor-color-nothing "" st)
  (dotimes (i (first dim-tab))
    (princ #\Space st))
  (vis-pos-color status-pos cursor-color-nothing "" st))


(defmacro con-transp (forma)
  "Consulta transparente, evita que PT-VIS haga seguimiento de ella"
  `(prog2
       (setf consulta-no-vis nil)
       ,forma
     (setf consulta-no-vis t)))
      

(defmacro limit-seq (seq-form limite)
  "Devuelve la sequencia SEQ limitada a LIMITE elementos o menos si no los alcanza y decrementa límite convenientemente"
  `(let* ((seq ,seq-form)
	  (num (min ,limite (length seq))))
     (prog1
	 (subseq seq 0 num)
       (setf ,limite (- ,limite num)))))

(defun print-nom-ag (&optional list-chocados (st *standard-output*))
  "Imprime nombres de agentes en la barra de estado."
  (clear-status st)
  (let ((limite (first dim-tab)))    ; Posiciones libres en status
    (dotimes (i num-jug-act)
      (let* ((jug (1+ i))
	     (agent-name (aref info-ag-jug jug 0))
	     (ene-jug (second (con-transp (envio:consulta-est-jug jug)))))
	(vis-pos-color nil 
		       (col-jug jug)
		       (limit-seq (format nil "~D" jug) limite)
		       st)
	;(format t "l:~S " limite)
	(if (member jug list-chocados)
	    (progn
	      (vis-pos-color nil color-end char-end st)
	      (setf limite (1- limite))
	      (vis-pos-color nil 
			     (col-jug jug)
			     (limit-seq (format nil "~C:~A:~4,'0D "
						(car-jug jug)
						agent-name
						ene-jug)
					limite)
			     st))
	  (format st (limit-seq (format nil ":~C:~A:~4,'0D "
					(car-jug jug)
					agent-name
					ene-jug)
				limite)))))))

(defun vis (tab &optional (st *standard-output*))
  "Visualiza el tablero TAB utlizando caracteres. TAB ha de ser una matriz con un tablero plantron (ver campo mat de pt-base:pt)."
  (set-window-cursor-position 1 1 st)
  (dotimes (y (array-dimension tab 1) t)
    (dotimes (x (array-dimension tab 0) (terpri st))
      (let ((elem (aref tab x y)))
	(if elem
	    (progn
	      (set-window-cursor-color (col-jug elem) st)
	      (princ (car-jug elem) st))
	  (princ " " st)))))
  ; Imprimimos versión de pt-vis
  (set-window-cursor-position 1 1 st)
  (format st "~A-~A" 
	  (package-name (symbol-package 'version))
	  (version))
  ; Guardamos e imprimimos leyenda
  (setf leyenda (envio:consulta-leyenda))
  (set-title-bar leyenda st)
  ; Imprimimos los nombres de los agentes
  (print-nom-ag nil st)
  ; Luego nos posicionamos en la línea de estado
  (vis-pos-color status-pos cursor-color-nothing "" st)
  (force-output st)
  t)

(defun vis-jug (jug pos &optional (st *standard-output*))
  "Visualiza el jugador JUG del tablero en la posición POS utlizando caracteres y colores."
  (vis-pos-color pos
		 (col-jug jug)
		 (car-jug jug)
		 st)
  t)

(defun vis-jug-end (est &optional (st *standard-output*))
  "Visualiza la finalización del juego, un ! en el choque. EST indica el estado de finalización del juego, ver función PT-BASE:CHECK-SET-INC-POS."
  ; Escribimos caracteres de final
  (dotimes (i num-jug-gdir)
    (let* ((jug (1+ i))
	   (posxy (con-transp (envio:consulta-pos-jug jug))))
      (if (member jug est)
	  (vis-pos-color posxy color-end char-end st)
	(vis-pos-color posxy
		       (col-jug jug)
		       (car-jug jug)
		       st))))
  ; Imprimimos los nombres de los agentes
  (print-nom-ag est st)
  ; Nos posicionamos en la línea de estado
  (vis-pos-color status-pos cursor-color-nothing "" st)
  (set-cursor-on st)
  (force-output st))

(defun vis-pos (pos &optional (st *standard-output*))
  "Visualiza (borra) una determinada posición POS del tablero PlanTron."
  (vis-jug (con-transp (envio:consulta-pos-tab pos)) pos st))

;; dir-inc, dir2dim-inc, dir2val-inc y pos-inc-dir
;; Extraidas de pt-base
;; ----------------------

(defconstant dir-inc
  (make-array '(4 2) :initial-contents
	      '((1 -1)    ; Norte
		(0 +1)    ; Este
		(1 +1)    ; Sur
		(0 -1)))  ; Oeste
  "Array para transformar de dirección a campo dir-jug de pt")

(defmacro dir2dim-inc (dir)
  `(aref dir-inc ,dir 0))

(defmacro dir2val-inc (dir)
  `(aref dir-inc ,dir 1))

(defun pos-inc-dir (pos dir)
  "Calcula la posición resultante de desplazarse en la direcciónb DIR respecto a POS."
  (let ((pos-inc (copy-list pos)))
    (setf (nth (dir2dim-inc dir) pos-inc)
	  (+ (nth (dir2dim-inc dir) pos-inc)
	     (dir2val-inc dir)))
    pos-inc))

(defun pinta-seg (seg &optional (st *standard-output*))
  "Pinta un seguimiento."
  (prog1
      (case (first seg)
	(cposj (vis-pos-color (fourth seg) (col-jug (second seg)) #\: st) seg)
	(cdirj (vis-pos-color (fourth seg) (col-jug (second seg)) #\? st) seg)
	(cpost
	 (if (fourth seg)
	     (setf (third seg) (pos-inc-dir (third seg) (fourth seg))))
         ; Evitar que posición a pintar (third seg) se salga del tablero
	 (let ((posx (first (third seg)))
	       (posy (second (third seg))))
	   (if (and (>= posx 0)
		    (>= posy 0)
		    (<= posx (1- (first dim-tab)))
		    (<= posy (1- (second dim-tab))))
	       (progn
		 (vis-pos-color (third seg)
				(col-jug (second seg))
				(case (fourth seg)
				  (0 #\^)
				  (1 #\>)
				  (2 #\v)
				  (3 #\<)
				  (otherwise #\,))
				st)
		 seg))))
	(cpunt (vis-pos-color (third seg) (col-jug (second seg)) #\- st) seg))
    (force-output st)))

(defun borra-segs (&optional (st *standard-output*))
  "Borra los seguimientos actualmente pintados."
  (if lis-con-seg
      (progn
	(if *step*
	    (progn
	      (force-output st)
	      (format t "Pulse return para borrar seguimientos")
	      (force-output)
	      (read-char)))
	(dolist (elm lis-con-seg)
	  (case (first elm)
	    ((cposj cdirj) (vis-jug (third elm) (fourth elm) st))
	    ((cpost cpunt) (vis-pos (third elm) *pt-vis-stream*))))
	(if *step*
	    (progn
	      (force-output st)
	      (format t "Pulse return para continuar")
	      (force-output)
	      (read-char)))
	(setf lis-con-seg nil))))

(defun borra-pinta-seg (&optional (st *standard-output*))
  "No usada..Borra los seguimientos actualmente pintados y pinta los nuevos."
  ; Borramos consultas las actualmente pintadas
  (dolist (elm lis-con-pin)
    (case (first elm)
      ((cposj cdirj) (vis-jug (third elm) (fourth elm) st))
      (cpost (vis-pos (third elm) *pt-vis-stream*))))
  ; Pintamos las consultas nuevas
  (dolist (elm (reverse lis-con-seg))
    (case (first elm)
      (cposj (vis-pos-color (fourth elm) (col-jug (second elm)) #\: st))
      (cdirj (vis-pos-color (fourth elm) (col-jug (second elm)) #\? st))
      (cpost
       (if (fourth elm)
	   (setf (third elm) (pos-inc-dir (third elm) (fourth elm))))
       (vis-pos-color (third elm)
		      (col-jug (second elm))
		      (case (fourth elm)
			(0 #\^)
			(1 #\>)
			(2 #\v)
			(3 #\<)
			(otherwise #\,))
			st))))
  ; Actualizamos listas
  (setf lis-con-pin lis-con-seg)
  (setf lis-con-seg nil))


;;
;; Interface de consultas envio ==> agente-visualización
;; -----------------------------------------------------

(defun ini-vis (jug tipo-juego)
  "Función genérica de inicialización de visualizador."
  (declare (ignore jug))
  ; Asigna variables globales
  (setf num-jug-gdir (getf tipo-juego :num-jug-gdir))
  (setf num-jug-act (getf tipo-juego :num-jug-act))
  (setf num-jug-pas (getf tipo-juego :num-jug-pas))
  (setf sleep-cs (getf tipo-juego :sleep-cs))
  (setf dim-tab (getf tipo-juego :dim-tab))
  (setf status-pos (calc-status-pos))
  (setf msg-jug (make-array num-jug-act :initial-element ""))
  (setf lis-con-seg nil
	lis-con-pin nil)

  ; Inicializa contador de movimientos para act. barra estado
  (setf int-ref-est (* *int-ref-est-jug* num-jug-gdir)
	cont-ref-est int-ref-est)

  ; Inicializa el array de nombres y documentación de jugadores
  (setf info-ag-jug (make-array (list (1+ num-jug-act) 2) :initial-element nil))
  (dotimes (i num-jug-act)
    (let* ((jug (1+ i))
	   (lis-keys (envio:consulta-lis-key-jug jug)))
      (setf (aref info-ag-jug jug 0) (getf lis-keys :name)
	    (aref info-ag-jug jug 1) (getf lis-keys :documentation))))


  ; Inicializa array de posiciones de jugadores
  (setf pos-jug-gdir (make-array num-jug-gdir :initial-element nil))
  (dotimes (i num-jug-gdir)
    (setf (aref pos-jug-gdir i) (con-transp (envio:consulta-pos-jug (1+ i)))))

  ; Inicializa array de caracteres y colores
  (setf elm-mat-char-color (make-array (list (+ num-jug-act 2) 2) :initial-element nil))
  (setf (aref elm-mat-char-color 0 0) char-obs
	(aref elm-mat-char-color 0 1) color-obs)
  (dotimes (i num-jug-gdir)
    (let ((ind (1+ i)))
      (setf (aref elm-mat-char-color ind 0) (aref color-char-jug (rem i 2) 0)
	    (aref elm-mat-char-color ind 1) (aref color-char-jug (rem i 2) 1))))
  (dotimes (i (- num-jug-act num-jug-gdir))
    (let ((ind (+ i num-jug-gdir 1)))
      (setf (aref elm-mat-char-color ind 0) (aref color-char-pla (rem i 2) 0)
	    (aref elm-mat-char-color ind 1) (aref color-char-pla (rem i 2) 1))))
  (setf (aref elm-mat-char-color (1+ num-jug-act) 0) char-end
	(aref elm-mat-char-color (1+ num-jug-act) 1) color-end)

  ; Visualiza mensajes previos
  (vis-prev)
  ; Visualiza tablero completo
  (set-cursor-off *pt-vis-stream*)
  (vis (envio:consulta-tab) *pt-vis-stream*)
  t)

(defun fin-vis (jug est)
  "Función de finalización de partida."
  (declare (ignore jug))
  ;(borra-pinta-seg *pt-vis-stream*)
  (borra-segs *pt-vis-stream*)
  (vis-jug-end est *pt-vis-stream*)
  (sleep (ceiling (* 12 sleep-cs) 100))
  ;(setf num-jug-act nil)  ; Para indicar paso a inactivo FIXME
  t)

(defun fin-vis0 (jug est)
  "Función de finalización de partida para agente vis-bajo."
  (vis (envio:consulta-tab) *pt-vis-stream*)
  (fin-vis jug est))

(defun seg-pos-vis (jug pos)
  "Función de seguimiento de cambio de posición de jugador."
  (if (eq jug num-jug-gdir)
      ; Borrar seguimientos con el último jugador FIXME
      ;(borra-pinta-seg *pt-vis-stream*))
      (borra-segs *pt-vis-stream*))
  (vis-jug jug pos *pt-vis-stream*)
  (if (<= jug num-jug-gdir)
      (setf (aref pos-jug-gdir (1- jug)) pos))
  ; Control refresco barra estado
  (if (zerop cont-ref-est)
      (progn
	(setf cont-ref-est int-ref-est)
	(print-nom-ag nil *pt-vis-stream*))
    (setf cont-ref-est (1- cont-ref-est)))
  (force-output *pt-vis-stream*)
  t)

(defun seg-dir-vis (jug dir)
  "Función de seguimiento de cambio de dirección de jugador. No implementada."
  (declare (ignore jug dir))
  nil)

(defmacro pinta-push-seg (lis-seg)
  "Pinta y apila seguimiento para luego borrar."
  `(let ((seg (pinta-seg ,lis-seg *pt-vis-stream*)))
     (if seg
	 (push seg lis-con-seg))))

(defun seg-cposj-vis  (jug-pet jug)
  "Función de seguimiento de consultas de posición de jugador."
  (if consulta-no-vis
      (pinta-push-seg (list 'cposj jug-pet jug (pos-jug jug)))))

(defun seg-cdirj-vis  (jug-pet jug)
  "Función de seguimiento de consultas de dirección de jugador."
  (if consulta-no-vis
      (pinta-push-seg (list 'cdirj jug-pet jug (pos-jug jug)))))

(defun seg-cpost-vis  (jug-pet pos dir)
  "Función de seguimiento de consultas de posición de tablero."
  (if consulta-no-vis
      (pinta-push-seg  (list 'cpost jug-pet pos dir))))

(defun rec-mensaje (jug msg)
  "Recibe el envío de un mensaje y lo imprime (en title-bar)."
  (let ((msg-tot leyenda))
;;    (if (and lis-ags-jug jug)
;;	(progn
	  (setf (aref msg-jug (1- jug)) msg)
	  (dotimes (i num-jug-act)
	    (if (string/= "" (aref msg-jug i))
		(setf msg-tot
		      (concatenate 'string msg-tot
				   (format nil "| ~D: ~A"
					   (1+ i)
					   (aref msg-jug i))))))
	  (set-title-bar msg-tot *pt-vis-stream*)))


;;
;; Interface de consultas agentes-jugadores ==> agente-visualización
;; -----------------------------------------------------------------

;; FIXME hacerlo pasar por envio
(defun envio-refresco-tab ()
  "Solicita refresco del tablero."
  (vis (envio:consulta-tab) *pt-vis-stream*)
  t)

;; FIXME hacerlo pasar por envio FIXME ¿chequeo activo con num-jug-act?
(defun envio-punteo-lis (lispos)
  "Solicita pintar punteo con '-' la lista de posiciones LISPOS."
  (if num-jug-act
      (dolist (pos lispos)
	(pinta-push-seg (list 'cpunt nil pos)))))

